/*****************************************************************
 *              String functions header v1.6                     *
 *                                                               *
 * 2006 Shengalts Aleksander aka Instructor (Shengalts@mail.ru)  *
 *                                                               *
 *                                                               *
 *Functions:                                                     *
 * WordFind, StrReplace, GetOptions                              *
 *                                                               *
 *Usage:                                                         *
 *#define WordFind                                               *
 *#define StrReplace                                             *
 *#define GetOptions                                             *
 *#include "StrFunc.h"                                           *
 *****************************************************************/

#ifndef _STRFUNC_
#define _STRFUNC_ 


/********************************************************************
 *
 *  WordFind
 *
 *String manipulation function.
 *
 *[out] char *pResult   -output
 *[in]  char *pText     -text
 *[in]  char *pDelim    -delimiter
 *[in]  int nNumber     -number of the delimiter/word if positive
 *                       search from beginning if negative from end,
 *                       if nNumber==0 then returns sum of delimiters/words
 *[in]  char *pOption   -">"     all text before founded delimiter
 *                       "<"     all text after founded delimiter
 *                       "<>"    deletes delimiter
 *                       "*"     text between delimiters (word)
 *                       "*>"    all text after founded word
 *                       ">*"    word and all text after founded word
 *                       "<*"    all text before founded word
 *                       "*<"    word and all text before founded word
 *                       "<*>"   deletes word and neighbouring delimiter
 *                       "/word" finds "word" and returns it number
 *                               if nNumber<0 then nRes=(0 - nRes)
 *[in]  BOOL bSensitive  -TRUE   case sensitive
 *                        FALSE  case insensitive
 *
 *Returns (nRes):  sum of delimiters/words if nNumber==0
 *                 0  no errors
 *                 -1 syntax error or pDelim is empty
 *                 -2 no delimiters found
 *                 -3 no such delimiter/word number
 *
 *Defines:
 * #define WordFind                 //insert WordFind function
 * #define WordFind_UNMINUS         //nNumber only positive (uses for minimize program size)
 * #define WordFind_UNPLUS          //nNumber only negative (uses for minimize program size)
 *
 *Examples:
 *  WordFind(szRes, "A||B||C", "||", 1, ">", TRUE);       //szRes == B||C
 *  WordFind(szRes, "A||B||C", "b", 1, ">", FALSE);       //szRes == ||C
 *  WordFind(szRes, "A||B||C", "||", 1, "<", TRUE);       //szRes == A
 *  WordFind(szRes, "A||B||C", "||", 2, "<", TRUE);       //szRes == A||B
 *  WordFind(szRes, "A||B||C", "||", -1, "<", TRUE);      //szRes == A||B
 *  WordFind(szRes, "A||B||C", "||", -2, "<>", TRUE);     //szRes == AB||C
 *  WordFind(szRes, "A||B||C", "||", 1, "*", TRUE);       //szRes == A
 *  WordFind(szRes, "A||B||C", "||", 2, "*", TRUE);       //szRes == B
 *  WordFind(szRes, "A||B||C", "||", -1, "*", TRUE);      //szRes == C
 *  WordFind(szRes, "A||B||C", "||", 2, "*>", TRUE);      //szRes == ||C
 *  WordFind(szRes, "A||B||C", "||", 2, ">*", TRUE);      //szRes == B||C
 *  WordFind(szRes, "A||B||C", "||", 2, "<*", TRUE);      //szRes == A||
 *  WordFind(szRes, "A||B||C", "||", 2, "*<", TRUE);      //szRes == A||B
 *  WordFind(szRes, "A||B||C", "||", 2, "<*>", TRUE);     //szRes == A||C
 *  nRes=WordFind(NULL, "A||B||C", "||", 0, ">", TRUE);  //nRes == 2
 *  nRes=WordFind(NULL, "A||B||C", "||", 0, "*", TRUE);  //nRes == 3
 *  nRes=WordFind(szRes, "A||B||C", "||", 1, "/B", TRUE); //nRes == 2
 ********************************************************************/
#ifdef WordFind
#undef WordFind
int WordFind(char *pResult, char *pText, char *pDelim, int nNumber, char *pOption, BOOL bSensitive)
{
	char *pWordStart=pText;
	char *pWordEnd=pText;
	char *pDelimStart;
	char *pDelimEnd;
	BOOL bFound=FALSE;
	BOOL bRow=TRUE;
	int nOption=0;
	int nCurStrNumber=0;
	int nCurWordNumber=0;
	int nCurWordNumberOp10=0;
	int nDelimLen=lstrlen(pDelim);

	if (!lstrcmp(pOption, ">")) nOption=1;
	else if (!lstrcmp(pOption, "<")) nOption=2;
	else if (!lstrcmp(pOption, "<>")) nOption=3;
	else if (!lstrcmp(pOption, "*")) nOption=4;
	else if (!lstrcmp(pOption, "*>")) nOption=5;
	else if (!lstrcmp(pOption, ">*")) nOption=6;
	else if (!lstrcmp(pOption, "<*")) nOption=7;
	else if (!lstrcmp(pOption, "*<")) nOption=8;
	else if (!lstrcmp(pOption, "<*>")) nOption=9;
	else if (pOption[0] == '/')
	{
		nOption=10;
		++pOption;
	}

	if (nDelimLen == 0 || nOption == 0)
	{
		lstrcpy(pResult, pText);
		return -1;
	}

	#ifndef WordFind_UNPLUS
	if (nNumber >= 0)
	{
		while (*pWordEnd)
		{
			for (pDelimStart=pDelim;
				(*pWordEnd) &&\
				((bSensitive == TRUE && *pWordEnd == *pDelimStart) ||\
				(bSensitive == FALSE && (char)CharUpper((char*)(unsigned)(unsigned char)*pWordEnd) == (char)CharUpper((char*)(unsigned)(unsigned char)*pDelimStart)));
				++pWordEnd)
			{
				if (!*++pDelimStart)
				{
					if (nOption > 3)
					{
						if (bRow == FALSE && (++nCurWordNumber == nNumber || nOption == 10))
						{
							if (nOption == 4 || nOption == 10)
								lstrcpyn(pResult, pWordStart, pWordEnd - pWordStart - nDelimLen + 2);
							else if (nOption == 5)
								lstrcpy(pResult, pWordEnd - nDelimLen + 1);
							else if (nOption == 6)
								lstrcpy(pResult, pWordStart);
							else if (nOption == 7)
								lstrcpyn(pResult, pText, pWordStart - pText + 1);
							else if (nOption == 8)
								lstrcpyn(pResult, pText, pWordEnd - pText - nDelimLen + 2);
							else if (nOption == 9)
							{
								lstrcpyn(pResult, pText, pWordStart - pText + 1);
								lstrcat(pResult, pWordEnd + 1);
							}
							if (nOption != 10)
								return 0;

							if (((bSensitive == TRUE && !lstrcmp(pResult, pOption)) ||\
								(bSensitive == FALSE && !lstrcmpi(pResult, pOption))) &&\
								++nCurWordNumberOp10 == nNumber)
								return nCurWordNumber;
						}
						pWordStart=pWordEnd + 1;
						bRow=TRUE;
					}
					else if (++nCurStrNumber == nNumber)
					{
						if (nOption == 1)
							lstrcpy(pResult, pWordEnd + 1);
						else if (nOption == 2)
							lstrcpyn(pResult, pText, pWordEnd - pText - nDelimLen + 2);
						else
						{
							lstrcpyn(pResult, pText, pWordEnd - pText - nDelimLen + 2);
							lstrcat(pResult, pWordEnd + 1);
						}
						return 0;
					}
					pDelimStart=pDelim;
					bFound=TRUE;
				}
			}
			if (*pWordEnd)
			{
				bRow=FALSE;
				++pWordEnd;
			}
		}
		if (bFound == TRUE && nOption > 3 && bRow == FALSE && (++nCurWordNumber == nNumber || nOption == 10))
		{
			if (nOption == 4 || nOption == 6 || nOption == 10)
				lstrcpy(pResult, pWordStart);
			else if (nOption == 5)
				pResult[0]='\0';
			else if (nOption == 7)
				lstrcpyn(pResult, pText, pWordStart - pText + 1);
			else if (nOption == 8)
				lstrcpy(pResult, pText);
			else if (nOption == 9)
				lstrcpyn(pResult, pText, pWordStart - pText - nDelimLen + 1);
			if (nOption != 10)
				return 0;

			if (((bSensitive == TRUE && !lstrcmp(pResult, pOption)) ||\
				(bSensitive == FALSE && !lstrcmpi(pResult, pOption))) &&\
				++nCurWordNumberOp10 == nNumber)
				return nCurWordNumber;
		}
		if (nNumber == 0)
		{
			if (nOption <= 3)
				return nCurStrNumber;
			if (nOption >= 4 && nOption <= 9)
				return nCurWordNumber;
			return nCurWordNumberOp10;
		}
	}
	#endif

	#ifndef WordFind_UNMINUS
	else if (nNumber < 0)
	{
		pWordStart+=lstrlen(pText) - 1;
		pWordEnd=pWordStart + 1;
		pDelimEnd=pDelim + nDelimLen - 1;

		while (pWordStart >= pText)
		{
			for (pDelimStart=pDelimEnd;
				(pWordStart >= pText) &&\
				((bSensitive == TRUE && *pWordStart == *pDelimStart) ||\
				(bSensitive == FALSE && (char)CharUpper((char*)(unsigned)(unsigned char)*pWordStart) == (char)CharUpper((char*)(unsigned)(unsigned char)*pDelimStart)));
				--pWordStart)
			{
				if (--pDelimStart < pDelim)
				{
					if (nOption > 3)
					{
						if (bRow == FALSE && (--nCurWordNumber == nNumber || nOption == 10))
						{
							if (nOption == 4 || nOption == 10)
								lstrcpyn(pResult, pWordStart + nDelimLen, pWordEnd - pWordStart - nDelimLen + 1);
							else if (nOption == 5)
								lstrcpy(pResult, pWordEnd);
							else if (nOption == 6)
								lstrcpy(pResult, pWordStart + nDelimLen);
							else if (nOption == 7)
								lstrcpyn(pResult, pText, pWordStart - pText + nDelimLen + 1);
							else if (nOption == 8)
								lstrcpyn(pResult, pText, pWordEnd - pText + 1);
							else if (nOption == 9)
							{
								lstrcpyn(pResult, pText, pWordStart - pText + 1);
								lstrcat(pResult, pWordEnd);
							}
							if (nOption != 10)
								return 0;

							if (((bSensitive == TRUE && !lstrcmp(pResult, pOption)) ||\
								(bSensitive == FALSE && !lstrcmpi(pResult, pOption))) &&\
								--nCurWordNumberOp10 == nNumber)
								return (0 - nCurWordNumber);
						}
						pWordEnd=pWordStart;
						bRow=TRUE;
					}
					else if (--nCurStrNumber == nNumber)
					{
						if (nOption == 1)
							lstrcpy(pResult, pWordStart + nDelimLen);
						else if (nOption == 2)
							lstrcpyn(pResult, pText, pWordStart - pText + 1);
						else
						{
							lstrcpyn(pResult, pText, pWordStart - pText + 1);
							lstrcat(pResult, pWordStart + nDelimLen);
						}
						return 0;
					}
					pDelimStart=pDelimEnd;
					bFound=TRUE;
				}
			}
			if (pWordStart >= pText)
			{
				bRow=FALSE;
				--pWordStart;
			}
		}
		if (bFound == TRUE && nOption > 3 && bRow == FALSE && (--nCurWordNumber == nNumber || nOption == 10))
		{
			if (nOption == 4 || nOption == 8 || nOption == 10)
				lstrcpyn(pResult, pText, pWordEnd - pText + 1);
			else if (nOption == 5)
				lstrcpy(pResult, pWordEnd);
			else if (nOption == 6)
				lstrcpy(pResult, pText);
			else if (nOption == 7)
				pResult[0]='\0';
			else if (nOption == 9)
				lstrcpy(pResult, pWordEnd + nDelimLen);
			if (nOption != 10)
				return 0;

			if (((bSensitive == TRUE && !lstrcmp(pResult, pOption)) ||\
				(bSensitive == FALSE && !lstrcmpi(pResult, pOption))) &&\
				--nCurWordNumberOp10 == nNumber)
				return (0 - nCurWordNumber);
		}
	}
	#endif

	lstrcpy(pResult, pText);
	if (bFound == TRUE)
		return -3;
	else
		return -2;
}
#endif

/********************************************************************
 *
 *  StrReplace
 *
 *Replace substring in string.
 *
 *[out] char *pResult    -output
 *[in]  char *pText      -text
 *[in]  char *pIt        -replace it
 *[in]  char *pWith      -replace with
 *[in]  BOOL bSensitive  -TRUE   case sensitive
 *                        FALSE  case insensitive
 *
 *Returns:  number of changes
 *
 *Defines:
 * #define StrReplace                 //insert StrReplace function
 *
 *Examples:
 *  StrReplace(szRes, "123||456||789||", "||", "+", TRUE);    //szRes == 123+456+789+
 *  StrReplace(szRes, "123||456||789||", "||", "", TRUE);     //szRes == 123456789
 ********************************************************************/
#ifdef StrReplace
#undef StrReplace
int StrReplace(char *pResult, char *pText, char *pIt, char *pWith, BOOL bSensitive)
{
	char *pText2=pText;
	char *pIt2=pIt;
	char *pResult2=pResult;
	int nItLen=lstrlen(pIt);
	int nWithLen=lstrlen(pWith);
	int nChanges=0;
	int i=0;

	while (*pText2)
		for (pIt2=pIt, pText2+=i, i=1;
			(*pText2) &&\
			((bSensitive == TRUE && *pText2 == *pIt2) ||\
			(bSensitive == FALSE && (char)CharUpper((char*)(unsigned)(unsigned char)*pText2) == (char)CharUpper((char*)(unsigned)(unsigned char)*pIt2)));
			++pText2)
			if (!*++pIt2)
			{
				lstrcpyn(pResult2, pText, pText2 - pText - nItLen + 2);
				lstrcat(pResult2, pWith);
				pResult2+=(pText2 - pText - nItLen + nWithLen + 1);
				pText=pText2 + 1;
				pIt2=pIt;
				++nChanges;
			}
	if (nChanges == 0)
		lstrcpy(pResult, pText);
	else
		lstrcat(pResult, pText);
	return nChanges;
}
#endif

/********************************************************************
 *
 *  GetOptions
 *
 *Gets option string from parameters line.
 *
 *[out] char *string     -option string
 *[in]  char *line       -parameters line
 *[in]  char *option     -option
 *[in]  BOOL bSensitive  -TRUE   case sensitive
 *                        FALSE  case insensitive
 *
 *Returns:  0 if option found
 *          1 if not found
 *
 *Defines:
 * #define GetOptions                 //insert GetOptions function
 *
 *Examples:
 *  GetOptions(szRes, "/a1=bb /a2=cc /a3=dd", "/a1=", TRUE);       //szRes == bb
 *  GetOptions(szRes, "/a1=bb /a2=\"cc\" /a3=dd", "/A2=", FALSE);   //szRes == cc
 *  GetOptions(szRes, "/a1=bb /a2=cc /a3=\"/dd\"", "/a3=", TRUE);  //szRes == /dd
 *  GetOptions(szRes, "/a1 /a2 /a3", "/a2", TRUE);                 //Returns: 0
 *  GetOptions(szRes, "/a1 /a2 /a3", "/a4", TRUE);                 //Returns: 1
 ********************************************************************/
#ifdef GetOptions
#undef GetOptions
int GetOptions(char *string, char *line, char *option, BOOL bSensitive)
{
	char chDelimiter='\0';
	char chQuote='\0';
	char *pTmp=NULL;
	int s=0;
	int nStart=0;
	int nTmp=0;

	chDelimiter=option[0];
	++option;

	for (; (line[s]); ++s)
	{
		if ((chQuote == '\0') && (line[s] != '\"') && (line[s] != '\'') && (line[s] != '`'))
		{
			if (line[s] == chDelimiter)
			{
				if (nStart)
				{
					end:
					while (line[--s] == ' ' || line[s] == '\t');

					if (((bSensitive == TRUE && line[nStart] == line[s]) ||\
						(bSensitive == FALSE && (char)CharUpper((char*)(unsigned)(unsigned char)line[nStart]) == (char)CharUpper((char*)(unsigned)(unsigned char)line[s]))) &&\
						((line[s] == '\"') || (line[s] == '\'') || (line[s] == '`')))
						++nStart, --s;

					for (; nStart <= s; ++nStart, ++nTmp)
						string[nTmp]=line[nStart];

					string[nTmp]='\0';
					return 0;
				}
				for (pTmp=option, ++s;
					 (line[s]) &&\
					 (*pTmp) &&\
					 ((bSensitive == TRUE && *pTmp == line[s]) ||\
					(bSensitive == FALSE && (char)CharUpper((char*)(unsigned)(unsigned char)*pTmp) == (char)CharUpper((char*)(unsigned)(unsigned char)line[s])));
					++s, ++pTmp);

				if ((!line[s]) && (!*pTmp))
				{
					string[0]='\0';
					return 0;
				}
				else if (!line[s])
				{
					string[0]='\0';
					return 1;
				}
				else if (!*pTmp)
					nStart=s--;
			}
		}
		else if (chQuote == '\0')
			chQuote=line[s];
		else if (line[s] == chQuote)
			chQuote='\0';
	}
	if (nStart) goto end;

	string[0]='\0';
	return 1;
}
#endif

#endif //_STRFUNC_