/*****************************************************************
 *                Registry NSIS plugin v2.4                      *
 *                                                               *
 * 2006 Shengalts Aleksander aka Instructor (Shengalts@mail.ru)  *
 *****************************************************************/


/* Comment-out and recompile */
#define REGISTRY_SEARCH        //Compile with search function
#define REGISTRY_KEY_EXISTS    //Compile with check key function
#define REGISTRY_READ          //Compile with read function
#define REGISTRY_WRITE         //Compile with write function
#define REGISTRY_READ_EXTRA    //Compile with extra read function
#define REGISTRY_WRITE_EXTRA   //Compile with extra write function
#define REGISTRY_CREATE        //Compile with create function
#define REGISTRY_DELETE        //Compile with delete key/value functions
#define REGISTRY_COPY_VALUE    //Compile with copy/move value functions
#define REGISTRY_COPY_KEY      //Compile with copy/move key functions
#define REGISTRY_BACKUP        //Compile with export/import functions
#define REGISTRY_CONVERT       //Compile with convertion functions




#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#ifdef POCKETPC
	#include <basetyps.h>
	#include "RAPI\rapi.h"
#endif

/* Defines */
#define MAX_PATHLEN        MAX_PATH+32
#define MAX_STRLEN         1024
#ifndef POCKETPC
	#define MAX_DATALEN        65536
#else
	#define MAX_DATALEN        4096
#endif

#ifndef REG_QWORD
	#define REG_QWORD 11
#endif
#define IDC_STATIC_DETAILS  1006
#define ALL_TYPES          -1

/* Include conversion functions */
#if defined REGISTRY_SEARCH || defined REGISTRY_WRITE || defined REGISTRY_READ || defined REGISTRY_READ_EXTRA || defined REGISTRY_CONVERT
	#if defined REGISTRY_WRITE || defined REGISTRY_READ_EXTRA
		#define xatoi
	#endif
	#if defined REGISTRY_CONVERT
		#define hex2str
	#endif
	#if defined REGISTRY_SEARCH || defined REGISTRY_READ || defined REGISTRY_READ_EXTRA || defined REGISTRY_CONVERT
		#define str2hex
	#endif
	#include "ConvFunc.h"
#endif

/* Include string functions */
#if defined REGISTRY_SEARCH || defined REGISTRY_BACKUP
	#define GetOptions
	#include "StrFunc.h"
#endif

/* Include private stack functions */
#if defined REGISTRY_SEARCH || defined REGISTRY_BACKUP
	#if defined REGISTRY_SEARCH || defined REGISTRY_BACKUP
		#define stack_insert
		#define stack_delete
		#define stack_clear
	#endif
	#include "StackFunc.h"

	stack_bilin *pStackElement=NULL, *pStackFirst=NULL, *pStackTop=NULL;
#endif

/* NSIS stack structure */
typedef struct _stack_t {
	struct _stack_t *next;
	char text[MAX_STRLEN];
} stack_t;

stack_t **g_stacktop;
char *g_variables;
unsigned int g_stringsize;

#define EXDLL_INIT()            \
{                               \
	g_stacktop=stacktop;      \
	g_variables=variables;    \
	g_stringsize=string_size; \
}

/* Global variables */
char szBuf[MAX_DATALEN]="";
char szBuf2[MAX_DATALEN]="";

#if defined REGISTRY_KEY_EXISTS || defined REGISTRY_READ || defined REGISTRY_WRITE ||\
    defined REGISTRY_READ_EXTRA || defined REGISTRY_WRITE_EXTRA || defined REGISTRY_CREATE ||\
    defined REGISTRY_DELETE || defined REGISTRY_COPY_VALUE || defined REGISTRY_COPY_KEY ||\
    defined REGISTRY_BACKUP || defined REGISTRY_CONVERT
	char szName[MAX_PATHLEN]="";
	char szPath[MAX_PATHLEN]="";
	char szKey[MAX_PATHLEN]="";
	char szValue[MAX_PATHLEN]="";
	unsigned char szString[MAX_DATALEN]="";
#endif

#ifdef REGISTRY_COPY_VALUE
	BOOL bMoveValue=FALSE;
#endif

#ifdef REGISTRY_COPY_KEY
	BOOL bMoveKey=FALSE;
#endif

#ifdef REGISTRY_SEARCH
	char szNameF[MAX_PATHLEN]="";
	char szPathF[MAX_PATHLEN]="";
	char szKeyF[MAX_PATHLEN]="";
	char szValueF[MAX_PATHLEN]="";
	unsigned char szStringF[MAX_DATALEN]="";
	char *pPathF=&szPathF[0];
	HKEY hKeyRootF=0;
	HKEY hKeyHandleF=0;
	HANDLE hSearchDetailsF=0;
	BOOL bAllNamesF=TRUE;
	BOOL bFindKeysF=TRUE;
	BOOL bFindValuesF=TRUE;
	BOOL bFindStringsF=TRUE;
	BOOL bGotoSubkeysF=TRUE;
	int nBannerF=0;
	int nFindTypeF=ALL_TYPES;
	int nGotoF=0;
	int nElementF=1;
	DWORD dwIndexF;
#endif

#ifdef POCKETPC
	wchar_t wszBuf[MAX_PATHLEN];
	wchar_t wszData[MAX_DATALEN/sizeof(wchar_t)];
#endif

/* Funtions prototypes and macros */
int TypeNameToTypeValue(char *name);
char* TypeValueToTypeName(DWORD type);
char* RegReadPath(HKEY *hKeyRoot, char **pPath);
void RegReadType(char *szOutput, char **pType, DWORD dwType, unsigned char *szString, DWORD dwSizeString);
void RegWriteType(HANDLE hFileWrite, char *szValueName, DWORD dwType, unsigned char *szString, DWORD dwSizeString);
void ReplaceEscapeSequence(char *pResult, char *pText);
LONG RegCopyKey(HKEY hSource, char *szPathSource, HKEY hTarget, char *szPathTarget);
int RegCopyValue(HKEY hSource, HKEY hTarget, char *szPathSource, char *pValueSource, char *szPathTarget, char *pValueTarget, BOOL bMove);
LONG RegDeleteSubKeys(HKEY hKeyRoot, LPCTSTR lpSubKey);
int RegDeleteKeyPath(HKEY hKeyRoot, char *szPath);
BOOL FileExists(char *fname);
void HexStrToHexData(char *pHexStr, unsigned char *pHexData, int *nHexLen);
#define SWAP_ENDIAN(x) (((x<<24)&0xFF000000)|((x<<8)&0xFF0000)|((x>>8)&0xFF00)|((x>>24)&0xFF))

LONG API_RegOpenKeyEx(HKEY hKey, LPCTSTR lpSubKey, DWORD ulOptions, REGSAM c, PHKEY phkResult);
LONG API_RegEnumKeyEx(HKEY hKey, DWORD dwIndex, LPTSTR lpName, LPDWORD lpcName, LPDWORD lpReserved, LPTSTR lpClass, LPDWORD lpcClass, PFILETIME lpftLastWriteTime);
LONG API_RegCreateKeyEx(HKEY hKey, LPCTSTR lpSubKey, DWORD Reserved, LPTSTR lpClass, DWORD dwOptions, REGSAM samDesired, LPSECURITY_ATTRIBUTES lpSecurityAttributes, PHKEY phkResult, LPDWORD lpdwDisposition);
LONG API_RegCloseKey(HKEY hKey);
LONG API_RegDeleteKey(HKEY hKey, LPCTSTR lpSubKey);
LONG API_RegEnumValue(HKEY hKey, DWORD dwIndex, LPTSTR lpValueName, LPDWORD lpcValueName, LPDWORD lpReserved, LPDWORD lpType, LPBYTE lpData, LPDWORD lpcbData);
LONG API_RegDeleteValue(HKEY hKey, LPCTSTR lpValueName);
LONG API_RegQueryValueEx(HKEY hKey, LPCTSTR lpValueName, LPDWORD lpReserved, LPDWORD lpType, LPBYTE lpData, LPDWORD lpcbData);
LONG API_RegSetValueEx(HKEY hKey, LPCTSTR lpValueName, DWORD Reserved, DWORD dwType, BYTE* lpData, DWORD cbData);

int popinteger(char *str);
int popstring(char *str);
void pushstring(const char *str);

/* NSIS functions code */
#ifdef REGISTRY_SEARCH
void __declspec(dllexport) _Open(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	int nKey=1;
	hKeyRootF=0;

	popstring(szPathF);
	popstring(szBuf);

	RegReadPath(&hKeyRootF, &pPathF);

	if (!GetOptions(szNameF, szBuf, "/N=", FALSE))
		bAllNamesF=FALSE;
	if (!GetOptions(szBuf2, szBuf, "/K=", FALSE) && *szBuf2 == '0')
		bFindKeysF=FALSE;
	if (!GetOptions(szBuf2, szBuf, "/V=", FALSE) && *szBuf2 == '0')
		bFindValuesF=FALSE;
	if (!GetOptions(szBuf2, szBuf, "/S=", FALSE) && *szBuf2 == '0')
		bFindStringsF=FALSE;
	if (!GetOptions(szBuf2, szBuf, "/G=", FALSE) && *szBuf2 == '0')
		bGotoSubkeysF=FALSE;
	if (!GetOptions(szBuf2, szBuf, "/B=", FALSE))
		if (*szBuf2 == '1')
		{
			nBannerF=1;
			hSearchDetailsF=GetDlgItem(FindWindowEx(hwndParent, NULL, "#32770", NULL), IDC_STATIC_DETAILS);
		}
		else if (*szBuf2 == '2')
			nBannerF=2;
	if (!GetOptions(szBuf2, szBuf, "/T=", FALSE))
		nFindTypeF=TypeNameToTypeValue(szBuf2);
	if (hKeyRootF == 0 || (bFindKeysF == FALSE && bFindValuesF == FALSE && bFindStringsF == FALSE) ||\
		API_RegOpenKeyEx(hKeyRootF, pPathF, 0, KEY_READ, &hKeyHandleF) != ERROR_SUCCESS)
	{
		pushstring("-1");
	}
	else
	{
		API_RegCloseKey(hKeyHandleF);
		pushstring("0");
		nGotoF=1;
	}
  }
}

void __declspec(dllexport) _Find(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pType;
	DWORD dwSizeKey;
	DWORD dwSizeValue;
	DWORD dwSizeString;
	DWORD dwType;
	int i;

	if (nGotoF == 4) goto EnumKey;
	else if (nGotoF == 3) goto EnumValue;
	else if (nGotoF == 2) goto OpenKey;
	else if (nGotoF != 1) goto End;

	stack_insert(&pStackElement, -1, &pStackFirst, &pStackTop);
	lstrcpy(pStackElement->string, pPathF);
	pPathF=&szPathF[0];

	while (nElementF != 0)
	{
		--nElementF;
		lstrcpy(pPathF, pStackTop->string);
		stack_delete(pStackTop, &pStackFirst, &pStackTop);

		if (nBannerF == 1) SendMessage(hSearchDetailsF, WM_SETTEXT, 0, (LPARAM)pPathF);
		else if (nBannerF == 2)
		{
			pushstring("BANNER");
			pushstring("");
			pushstring("");
			pushstring(pPathF);

			nGotoF=2;
			return;
		}

		OpenKey:
		if (API_RegOpenKeyEx(hKeyRootF, pPathF, 0, KEY_READ, &hKeyHandleF) != ERROR_SUCCESS) continue;

		dwIndexF=-1;
		if (bFindValuesF == FALSE && bFindStringsF == FALSE) goto EnumKey;

		EnumValue:
		++dwIndexF;
		dwSizeValue=MAX_PATHLEN;
		dwSizeString=MAX_DATALEN;
		szStringF[0]='\0';

		if (API_RegEnumValue(hKeyHandleF, dwIndexF, szValueF, &dwSizeValue, NULL, &dwType, szStringF, &dwSizeString) == ERROR_SUCCESS)
		{
			if ((bFindValuesF == TRUE && (bAllNamesF == TRUE || !lstrcmpi(szValueF, szNameF)) && (nFindTypeF == ALL_TYPES || nFindTypeF == (int)dwType)) ||\
				(bFindStringsF == TRUE && (bAllNamesF == TRUE || ((dwType == REG_SZ || dwType == REG_EXPAND_SZ) && !lstrcmpi(szStringF, szNameF))) && (nFindTypeF == ALL_TYPES || nFindTypeF == (int)dwType)))
			{
				RegReadType(szBuf, &pType, dwType, szStringF, dwSizeString);
				szBuf[MAX_STRLEN - 2]='\0';
				pushstring(pType);
				pushstring(szBuf);
				pushstring(szValueF);
				pushstring(pPathF);

				nGotoF=3;
				return;
			}
			else goto EnumValue;
		}

		dwIndexF=-1;

		EnumKey:
		if (bGotoSubkeysF != FALSE)
		{
			++dwIndexF;
			dwSizeKey=MAX_PATHLEN;

			if (API_RegEnumKeyEx(hKeyHandleF, dwIndexF, szKeyF, &dwSizeKey, NULL, NULL, NULL, NULL) == ERROR_SUCCESS)
			{
				lstrcpy(szBuf, pPathF);
				if (*pPathF) lstrcat(szBuf, "\\");
				lstrcat(szBuf, szKeyF);

				//stack_insert(&pStackElement, dwIndexF + 1, &pStackFirst, &pStackTop);
				//lstrcpy(pStackElement->string, szBuf);
				stack_insert(&pStackElement, -1, &pStackFirst, &pStackTop);
				lstrcpy(pStackElement->string, szBuf);

				++nElementF;

				if (bFindKeysF == TRUE && (bAllNamesF == TRUE || !lstrcmpi(szKeyF, szNameF)))
				{
					pushstring("REG_KEY");
					pushstring("");
					pushstring(szKeyF);
					pushstring(pPathF);

					nGotoF=4;
					return;
				}
				else goto EnumKey;
			}
		}
		API_RegCloseKey(hKeyHandleF);
	}
	hKeyHandleF=0;
	nGotoF=0;

	End:
	for(i=0; i < 4; ++i) pushstring("");
  }
}

void __declspec(dllexport) _Close(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  {
	if (hKeyHandleF) API_RegCloseKey(hKeyHandleF);
	if (nBannerF == 1) SendMessage(hSearchDetailsF, WM_SETTEXT, 0, (LPARAM)"");
	stack_clear(&pStackFirst, &pStackTop);
	hKeyHandleF=0;
	nGotoF=0;
  }
}
#endif //REGISTRY_SEARCH

#ifdef REGISTRY_KEY_EXISTS
void __declspec(dllexport) _KeyExists(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pPath=&szPath[0];
	HKEY hKeyRoot=0;
	HKEY hKeyHandle=0;
	int nError=-1;

	popstring(szPath);

	RegReadPath(&hKeyRoot, &pPath);

	if (API_RegOpenKeyEx(hKeyRoot, pPath, 0, KEY_READ, &hKeyHandle) == ERROR_SUCCESS)
	{
		nError=API_RegCloseKey(hKeyHandle);
	}
	pushstring((nError == ERROR_SUCCESS)?"0":"-1");
  }
}
#endif //REGISTRY_KEY_EXISTS

#ifdef REGISTRY_READ
void __declspec(dllexport) _Read(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pPath=&szPath[0];
	char *pType;
	HKEY hKeyRoot=0;
	HKEY hKeyHandle=0;
	DWORD dwSizeString=MAX_DATALEN;
	DWORD dwType;
	int nError=-1;
	szString[0]='\0';

	popstring(szPath);
	popstring(szName);

	RegReadPath(&hKeyRoot, &pPath);

	if (API_RegOpenKeyEx(hKeyRoot, pPath, 0, KEY_QUERY_VALUE, &hKeyHandle) == ERROR_SUCCESS)
	{
		nError=API_RegQueryValueEx(hKeyHandle, szName, NULL, &dwType, szString, &dwSizeString);

		API_RegCloseKey(hKeyHandle);

		if (nError == ERROR_SUCCESS)
		{
			RegReadType(szBuf, &pType, dwType, szString, dwSizeString);
			szBuf[MAX_STRLEN - 2]='\0';
		}
	}
	pushstring((nError == ERROR_SUCCESS)?pType:"");
	pushstring((nError == ERROR_SUCCESS)?szBuf:"");
  }
}
#endif //REGISTRY_READ

#ifdef REGISTRY_WRITE
void __declspec(dllexport) _Write(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pBuf=&szBuf[0];
	char *pPath=&szPath[0];
	HKEY hKeyRoot=0;
	HKEY hKeyHandle=0;
	DWORD dwValue;
	DWORD dwSizeString;
	DWORD dwType;
	DWORD dwDisposition;
	int nError=-1;

	popstring(szPath);
	popstring(szName);
	popstring(szBuf);
	popstring(szBuf2);

	szBuf[MAX_STRLEN - 2]='\0';
	RegReadPath(&hKeyRoot, &pPath);

	if ((dwType=TypeNameToTypeValue(szBuf2)) != -1 && API_RegCreateKeyEx(hKeyRoot, pPath, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKeyHandle, &dwDisposition) == ERROR_SUCCESS)
	{
		if (dwType == REG_BINARY ||\
			dwType == REG_NONE ||\
			dwType == REG_LINK ||\
			dwType == REG_RESOURCE_LIST ||\
			dwType == REG_FULL_RESOURCE_DESCRIPTOR ||\
			dwType == REG_RESOURCE_REQUIREMENTS_LIST ||\
			dwType == REG_QWORD)
		{
			HexStrToHexData(szBuf, (unsigned char *)szBuf2, &dwSizeString);
			nError=API_RegSetValueEx(hKeyHandle, szName, 0, dwType, (unsigned char *)szBuf2, dwSizeString);
		}
		else if (dwType == REG_DWORD ||\
			dwType == REG_DWORD_BIG_ENDIAN)
		{
			if (dwType == REG_DWORD) dwValue=xatoi(szBuf);
			else dwValue=SWAP_ENDIAN(xatoi(szBuf));
			nError=API_RegSetValueEx(hKeyHandle, szName, 0, dwType, (unsigned char *)&dwValue, sizeof(DWORD));
		}
		else if (dwType == REG_MULTI_SZ)
		{
			while (*pBuf)
			{
				if (*pBuf == '\n') *pBuf='\0';
				++pBuf;
			}
			*++pBuf='\0';

			nError=API_RegSetValueEx(hKeyHandle, szName, 0, REG_MULTI_SZ, szBuf, (pBuf - szBuf) + 1);
		}
		else if (dwType == REG_EXPAND_SZ ||\
			dwType == REG_SZ)
			nError=API_RegSetValueEx(hKeyHandle, szName, 0, dwType, szBuf, lstrlen(szBuf) + 1);

		API_RegCloseKey(hKeyHandle);
	}
	pushstring((nError == ERROR_SUCCESS)?"0":"-1");
  }
}
#endif //REGISTRY_WRITE

#ifdef REGISTRY_READ_EXTRA
void __declspec(dllexport) _ReadExtra(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pBuf=&szBuf[0];
	char *pPath=&szPath[0];
	char *pType;
	HKEY hKeyRoot=0;
	HKEY hKeyHandle=0;
	DWORD dwSizeString=MAX_DATALEN;
	DWORD dwType;
	int nError=-1;
	int nPointer;
	szString[0]='\0';

	popstring(szPath);
	popstring(szName);
	nPointer=popinteger(szBuf);

	RegReadPath(&hKeyRoot, &pPath);

	if (API_RegOpenKeyEx(hKeyRoot, pPath, 0, KEY_QUERY_VALUE, &hKeyHandle) == ERROR_SUCCESS)
	{
		nError=API_RegQueryValueEx(hKeyHandle, szName, NULL, &dwType, szString, &dwSizeString);

		API_RegCloseKey(hKeyHandle);

		if (nError == ERROR_SUCCESS)
		{
			RegReadType(szBuf, &pType, dwType, szString, dwSizeString);

			if (dwType == REG_BINARY ||\
				dwType == REG_NONE ||\
				dwType == REG_LINK ||\
				dwType == REG_RESOURCE_LIST ||\
				dwType == REG_FULL_RESOURCE_DESCRIPTOR ||\
				dwType == REG_RESOURCE_REQUIREMENTS_LIST ||\
				dwType == REG_QWORD)
				nPointer*=2;
			if (nPointer > 0)
			{
				if (nPointer < lstrlen(szBuf)) pBuf+=nPointer;
				else pBuf[0]='\0';
			}
			else if (nPointer < 0 && (0 - nPointer) < lstrlen(szBuf))
				pBuf+=(lstrlen(szBuf) + nPointer);

			pBuf[MAX_STRLEN - 2]='\0';
		}
	}
	pushstring((nError == ERROR_SUCCESS)?pType:"");
	pushstring((nError == ERROR_SUCCESS)?pBuf:"");
  }
}
#endif //REGISTRY_READ_EXTRA

#ifdef REGISTRY_WRITE_EXTRA
void __declspec(dllexport) _WriteExtra(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pBuf=&szBuf[0];
	char *pPath=&szPath[0];
	HKEY hKeyRoot=0;
	HKEY hKeyHandle=0;
	DWORD dwSizeString=MAX_DATALEN;
	DWORD dwSizeStringExtra;
	DWORD dwType;
	int nError=-1;
	int a,b,c;
	szString[0]='\0';

	popstring(szPath);
	popstring(szName);
	popstring(szBuf);

	szBuf[MAX_STRLEN - 2]='\0';
	RegReadPath(&hKeyRoot, &pPath);

	if (API_RegOpenKeyEx(hKeyRoot, pPath, 0, KEY_QUERY_VALUE|KEY_SET_VALUE, &hKeyHandle) == ERROR_SUCCESS)
	{
		nError=API_RegQueryValueEx(hKeyHandle, szName, NULL, &dwType, szString, &dwSizeString);

		if (nError == ERROR_SUCCESS && *szBuf)
		{
			if (dwType == REG_BINARY ||\
				dwType == REG_NONE ||\
				dwType == REG_LINK ||\
				dwType == REG_RESOURCE_LIST ||\
				dwType == REG_FULL_RESOURCE_DESCRIPTOR ||\
				dwType == REG_RESOURCE_REQUIREMENTS_LIST ||\
				dwType == REG_QWORD)
			{
				HexStrToHexData(szBuf, (unsigned char *)szBuf2, &dwSizeStringExtra);
				for (a=0, b=dwSizeString, c=dwSizeStringExtra; a < c; szString[b++]=szBuf2[a++]);
				nError=API_RegSetValueEx(hKeyHandle, szName, 0, dwType, szString, (dwSizeString + c));
			}
			else if (dwType == REG_MULTI_SZ)
			{
				while (*pBuf)
				{
					if (*pBuf == '\n') *pBuf='\0';
					++pBuf;
				}
				*++pBuf='\0';

				for (a=0, b=dwSizeString - 1; a < 2 && b >= 0 && !szString[b]; ++a, --b);
				for (a=0, ++b, c=(pBuf - szBuf) + 1; a < c; szString[b++]=szBuf[a++]);
				nError=API_RegSetValueEx(hKeyHandle, szName, 0, REG_MULTI_SZ, szString, b);
			}
			else if (dwType == REG_EXPAND_SZ ||\
				dwType == REG_SZ)
			{
				lstrcat(szString, szBuf);
				nError=API_RegSetValueEx(hKeyHandle, szName, 0, dwType, szString, lstrlen(szString) + 1);
			}
		}
		API_RegCloseKey(hKeyHandle);
	}
	pushstring((nError == ERROR_SUCCESS)?"0":"-1");
  }
}
#endif //REGISTRY_WRITE_EXTRA

#ifdef REGISTRY_CREATE
void __declspec(dllexport) _CreateKey(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pPath=&szPath[0];
	HKEY hKeyRoot=0;
	HKEY hKeyHandle=0;
	DWORD dwDisposition;

	popstring(szPath);

	RegReadPath(&hKeyRoot, &pPath);

	if (API_RegCreateKeyEx(hKeyRoot, pPath, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKeyHandle, &dwDisposition) == ERROR_SUCCESS)
	{
		API_RegCloseKey(hKeyHandle);

		pushstring((dwDisposition == REG_CREATED_NEW_KEY)?"0":"1");
		return;
	}
	pushstring("-1");
  }
}
#endif //REGISTRY_CREATE

#ifdef REGISTRY_DELETE
void __declspec(dllexport) _DeleteValue(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pPath=&szPath[0];
	HKEY hKeyRoot=0;
	HKEY hKeyHandle=0;
	int nError=-1;

	popstring(szPath);
	popstring(szName);

	RegReadPath(&hKeyRoot, &pPath);

	if (API_RegOpenKeyEx(hKeyRoot, pPath, 0, KEY_SET_VALUE, &hKeyHandle) == ERROR_SUCCESS)
	{
		nError=API_RegDeleteValue(hKeyHandle, szName);

		API_RegCloseKey(hKeyHandle);
	}
	pushstring((nError == ERROR_SUCCESS)?"0":"-1");
  }
}

void __declspec(dllexport) _DeleteKey(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pPath=&szPath[0];
	HKEY hKeyRoot=0;
	int nError=-1;

	popstring(szPath);

	RegReadPath(&hKeyRoot, &pPath);

	nError=RegDeleteKeyPath(hKeyRoot, pPath);

	pushstring((nError == ERROR_SUCCESS)?"0":"-1");
  }
}

void __declspec(dllexport) _DeleteKeyEmpty(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pPath=&szPath[0];
	HKEY hKeyRoot=0;
	HKEY hKeyHandle=0;
	int dwSizeKey=MAX_PATHLEN;
	int dwSizeValue=MAX_PATHLEN;
	int dwType;
	int nError=-1;
	int nErrorKey=-1;
	int nErrorValue=-1;
	popstring(szPath);

	RegReadPath(&hKeyRoot, &pPath);

	if (API_RegOpenKeyEx(hKeyRoot, pPath, 0, KEY_READ, &hKeyHandle) == ERROR_SUCCESS)
	{
		if (API_RegEnumKeyEx(hKeyHandle, 0, szKey, &dwSizeKey, NULL, NULL, NULL, NULL) != ERROR_SUCCESS &&\
			API_RegEnumValue(hKeyHandle, 0, szValue, &dwSizeValue, NULL, &dwType, NULL, NULL) != ERROR_SUCCESS)
		{
			nError=ERROR_SUCCESS;
		}
		API_RegCloseKey(hKeyHandle);

		if (nError == ERROR_SUCCESS)
		{
			nError=RegDeleteKeyPath(hKeyRoot, pPath);
		}
	}
	pushstring((nError == ERROR_SUCCESS)?"0":"-1");
  }
}
#endif //REGISTRY_DELETE

#ifdef REGISTRY_COPY_VALUE
void __declspec(dllexport) _CopyValue(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pBuf=&szBuf[0];
	char *pPath=&szPath[0];
	DWORD dwSizeString=MAX_DATALEN;
	DWORD dwType;
	DWORD dwDisposition;
	HKEY hKeyRootSource=0;
	HKEY hKeyRootTarget=0;
	HKEY hSourceHandle;
	HKEY hTargetHandle;
	int nError=-1;
	szString[0]='\0';

	popstring(szPath);
	popstring(szName);
	popstring(szBuf);
	popstring(szBuf2);

	RegReadPath(&hKeyRootSource, &pPath);
	RegReadPath(&hKeyRootTarget, &pBuf);

	if (API_RegOpenKeyEx(hKeyRootSource, pPath, 0, KEY_QUERY_VALUE|KEY_SET_VALUE, &hSourceHandle) == ERROR_SUCCESS)
	{
		nError=API_RegQueryValueEx(hSourceHandle, szName, NULL, &dwType, szString, &dwSizeString);

		if (nError == ERROR_SUCCESS && (nError=API_RegCreateKeyEx(hKeyRootTarget, pBuf, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hTargetHandle, &dwDisposition)) == ERROR_SUCCESS)
		{
			nError=API_RegSetValueEx(hTargetHandle, szBuf2, 0, dwType, szString, dwSizeString);

			API_RegCloseKey(hTargetHandle);

			if (bMoveValue == TRUE && nError == ERROR_SUCCESS)
				nError=API_RegDeleteValue(hSourceHandle, szName);
		}
		API_RegCloseKey(hSourceHandle);
	}
	pushstring((nError == ERROR_SUCCESS)?"0":"-1");
  }
}

void __declspec(dllexport) _MoveValue(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
	bMoveValue=TRUE;
	_CopyValue(hwndParent, string_size, variables, stacktop);
	bMoveValue=FALSE;
}
#endif //REGISTRY_COPY_VALUE

#ifdef REGISTRY_COPY_KEY
void __declspec(dllexport) _CopyKey(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pBuf=&szBuf[0];
	char *pPath=&szPath[0];
	HKEY hKeyRoot=0;
	HKEY hKeyRootTarget=0;
	int nError;

	popstring(szPath);
	popstring(szBuf);

	RegReadPath(&hKeyRoot, &pPath);
	RegReadPath(&hKeyRootTarget, &pBuf);

	if ((nError=RegCopyKey(hKeyRoot, pPath, hKeyRootTarget, pBuf)) == ERROR_SUCCESS)
		if (bMoveKey == TRUE) nError=RegDeleteKeyPath(hKeyRoot, pPath);

	pushstring((nError == ERROR_SUCCESS)?"0":"-1");
  }
}

void __declspec(dllexport) _MoveKey(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
	bMoveKey=TRUE;
	_CopyKey(hwndParent, string_size, variables, stacktop);
	bMoveKey=FALSE;
}
#endif //REGISTRY_COPY_KEY

#ifdef REGISTRY_BACKUP
void __declspec(dllexport) _SaveKey(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	char *pPath=&szPath[0];
	const char *pRootName;
	HKEY hKeyRoot=0;
	HKEY hKeyHandle=0;
	DWORD dwSizeKey;
	DWORD dwSizeValue;
	DWORD dwSizeString;
	DWORD dwType;
	DWORD dwIndex;
	DWORD dwNumberOfBytesWritten;
	int nReplaceSubkeys=0;
	int nElement=1;
	BOOL bGotoSubkeys=TRUE;
	HANDLE hSearchDetails=0;
	HANDLE hFileWrite=0;

	popstring(szPath);
	popstring(szName);
	popstring(szBuf);

	pRootName=RegReadPath(&hKeyRoot, &pPath);

	if (!GetOptions(szBuf2, szBuf, "/G=", FALSE))
		if (*szBuf2 == '0') bGotoSubkeys=FALSE;
	if (!GetOptions(szBuf2, szBuf, "/B=", FALSE))
		if (*szBuf2 == '1') hSearchDetails=GetDlgItem(FindWindowEx(hwndParent, NULL, "#32770", NULL), IDC_STATIC_DETAILS);
	if (!GetOptions(szBuf2, szBuf, "/D=", FALSE))
		if (*szBuf2 == '1') nReplaceSubkeys=1;
		else if (*szBuf2 == '2') nReplaceSubkeys=2;

	if (API_RegOpenKeyEx(hKeyRoot, pPath, 0, KEY_READ, &hKeyHandle) != ERROR_SUCCESS)
	{
		pushstring("-1");
		return;
	}
	API_RegCloseKey(hKeyHandle);

	hFileWrite=CreateFile(szName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
	if (hFileWrite == INVALID_HANDLE_VALUE)
	{
		pushstring("-1");
		return;
	}
	WriteFile(hFileWrite, "REGEDIT4\r\n", lstrlen("REGEDIT4\r\n"), &dwNumberOfBytesWritten, NULL);

	stack_insert(&pStackElement, 1, &pStackFirst, &pStackTop);
	lstrcpy(pStackElement->string, pPath);
	pPath=&szPath[0];

	while (nElement != 0)
	{
		--nElement;
		lstrcpy(pPath, pStackTop->string);
		stack_delete(pStackTop, &pStackFirst, &pStackTop);

		if (hSearchDetails != 0) SendMessage(hSearchDetails, WM_SETTEXT, 0, (LPARAM)pPath);

		if (API_RegOpenKeyEx(hKeyRoot, pPath, 0, KEY_READ, &hKeyHandle) != ERROR_SUCCESS) continue;

		if (!*pPath)
			wsprintf(szBuf, "\r\n[%s]\r\n", pRootName);
		else
		{
			if (nReplaceSubkeys == 1 || nReplaceSubkeys == 2)
			{
				if (nReplaceSubkeys == 1) nReplaceSubkeys=0;
				wsprintf(szBuf, "\r\n[-%s\\%s]", pRootName, pPath);
				WriteFile(hFileWrite, szBuf, lstrlen(szBuf), &dwNumberOfBytesWritten, NULL);
			}
			wsprintf(szBuf, "\r\n[%s\\%s]\r\n", pRootName, pPath);
		}

		WriteFile(hFileWrite, szBuf, lstrlen(szBuf), &dwNumberOfBytesWritten, NULL);

		dwIndex=0;
		while (1)
		{
			dwSizeValue=MAX_PATHLEN;
			dwSizeString=MAX_DATALEN;
			szString[0]='\0';
			if (API_RegEnumValue(hKeyHandle, dwIndex++, szValue, &dwSizeValue, NULL, &dwType, szString, &dwSizeString) != ERROR_SUCCESS)
				break;

			RegWriteType(hFileWrite, szValue, dwType, szString, dwSizeString);
		}

		if (bGotoSubkeys != FALSE)
		{
			dwIndex=0;
			while (1)
			{
				dwSizeKey=MAX_PATHLEN;
				if (API_RegEnumKeyEx(hKeyHandle, dwIndex++, szKey, &dwSizeKey, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
					break;

				lstrcpy(szBuf, pPath);
				if (*pPath) lstrcat(szBuf, "\\");
				lstrcat(szBuf, szKey);

				stack_insert(&pStackElement, dwIndex, &pStackFirst, &pStackTop);
				lstrcpy(pStackElement->string, szBuf);

				++nElement;
			}
		}
		API_RegCloseKey(hKeyHandle);
	}
	CloseHandle(hFileWrite);
	stack_clear(&pStackFirst, &pStackTop);
	if (hSearchDetails != 0) SendMessage(hSearchDetails, WM_SETTEXT, 0, (LPARAM)"");
	pushstring("0");
  }
}

#ifndef POCKETPC
void __declspec(dllexport) _RestoreKey(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	STARTUPINFO si;
	PROCESS_INFORMATION pi;

	popstring(szName);

	if (FileExists(szName) == 0) goto Error;
	if (SearchPath(NULL, "regedit.exe", NULL, sizeof(szBuf2), szBuf2, 0) == 0) goto Error;
	wsprintf(szBuf, "%s /s \"%s\"", szBuf2, szName);

	GetStartupInfo(&si);

	if (CreateProcess(NULL, szBuf, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
	{
		pushstring("0");
		return;
	}

	Error:
	pushstring("-1");
  }
}
#endif //POCKETPC
#endif //REGISTRY_BACKUP

#ifdef REGISTRY_CONVERT
void __declspec(dllexport) _StrToHex(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	popstring(szBuf);

	str2hex((unsigned char *)szBuf, szBuf2, TRUE, lstrlen(szBuf));

	pushstring(szBuf2);
  }
}

void __declspec(dllexport) _HexToStr(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	popstring(szBuf);

	hex2str(szBuf, szBuf2);

	pushstring(szBuf2);
  }
}
#endif //REGISTRY_CONVERT

#ifdef POCKETPC
void __declspec(dllexport) _CERAPIINIT(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	if (CeRapiInit() == S_OK)
	{
		pushstring("0");
		return;
	}
	//Unable to initialize connection to device
	pushstring("-1");
  }
}

void __declspec(dllexport) _CERAPIINITEX(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
  EXDLL_INIT();
  {
	RAPIINIT rapiinit;

	rapiinit.cbSize=sizeof(RAPIINIT);
	rapiinit.heRapiInit=NULL;
	rapiinit.hrRapiInit=-1;

	if (CeRapiInitEx(&rapiinit) != S_OK)
	{
		//Unable to initialize connection to device
		pushstring("-1");
		return;
	}
	if (MsgWaitForMultipleObjects(1, &rapiinit.heRapiInit, FALSE, 3000, 0) == WAIT_TIMEOUT)
	{
		//Timed out waiting for device
		CeRapiUninit();
		pushstring("-2");
		return;
	}
	if (rapiinit.hrRapiInit != S_OK)
	{
		//Failed to connect to device
		CeRapiUninit();
		pushstring("-3");
		return;
	}
	pushstring("0");
  }
}

void __declspec(dllexport) _CERAPIUNINIT(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
	CeRapiUninit();
}
#endif //POCKETPC

void __declspec(dllexport) _Unload(HWND hwndParent, int string_size,
                                      char *variables, stack_t **stacktop)
{
}

BOOL WINAPI DllMain(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
#ifdef _DEBUG
	switch(ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
		printf("");
		break;
	case DLL_PROCESS_DETACH:
		printf("");
		break;
	case DLL_THREAD_ATTACH:
		printf("");
		break;
	case DLL_THREAD_DETACH:
		printf("");
		break;
	}
#endif

	return TRUE;
}


/* Functions */
#if defined REGISTRY_SEARCH || defined REGISTRY_KEY_EXISTS || defined REGISTRY_READ ||\
    defined REGISTRY_WRITE || defined REGISTRY_READ_EXTRA || defined REGISTRY_WRITE_EXTRA ||\
    defined REGISTRY_CREATE || defined REGISTRY_DELETE || defined REGISTRY_COPY_VALUE ||\
    defined REGISTRY_COPY_KEY || defined REGISTRY_BACKUP
// Function: "HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft" -> HKEY_LOCAL_MACHINE "SOFTWARE\Microsoft"
char* RegReadPath(HKEY *hKeyRoot, char **pPath)
{
	char *pKey=*pPath + lstrlen(*pPath);
	char *pKeyRoot=*pPath;

	while (*--pKey == '\\') *pKey='\0';

	for (pKey=*pPath; *pKey && *pKey != '\\'; ++pKey);

	if (*pKey == '\\')
	{
		*pKey='\0';
		++pKey;
	}
	*pPath=pKey;

	if (!lstrcmpi(pKeyRoot, "HKEY_CLASSES_ROOT") || !lstrcmpi(pKeyRoot, "HKCR"))
	{
		*hKeyRoot=HKEY_CLASSES_ROOT;
		return "HKEY_CLASSES_ROOT";
	}
	else if (!lstrcmpi(pKeyRoot, "HKEY_CURRENT_USER") || !lstrcmpi(pKeyRoot, "HKCU"))
	{
		*hKeyRoot=HKEY_CURRENT_USER;
		return "HKEY_CURRENT_USER";
	}
	else if (!lstrcmpi(pKeyRoot, "HKEY_LOCAL_MACHINE") || !lstrcmpi(pKeyRoot, "HKLM"))
	{
		*hKeyRoot=HKEY_LOCAL_MACHINE;
		return "HKEY_LOCAL_MACHINE";
	}
	else if (!lstrcmpi(pKeyRoot, "HKEY_USERS") || !lstrcmpi(pKeyRoot, "HKU"))
	{
		*hKeyRoot=HKEY_USERS;
		return "HKEY_USERS";
	}
	else if (!lstrcmpi(pKeyRoot, "HKEY_PERFORMANCE_DATA") || !lstrcmpi(pKeyRoot, "HKPD"))
	{
		*hKeyRoot=HKEY_PERFORMANCE_DATA;
		return "HKEY_PERFORMANCE_DATA";
	}
	else if (!lstrcmpi(pKeyRoot, "HKEY_CURRENT_CONFIG") || !lstrcmpi(pKeyRoot, "HKCC"))
	{
		*hKeyRoot=HKEY_CURRENT_CONFIG;
		return "HKEY_CURRENT_CONFIG";
	}
	else if (!lstrcmpi(pKeyRoot, "HKEY_DYN_DATA") || !lstrcmpi(pKeyRoot, "HKDD"))
	{
		*hKeyRoot=HKEY_DYN_DATA;
		return "HKEY_DYN_DATA";
	}
	else
	{
		*hKeyRoot=0;
		return "";
	}
}
#endif //REGISTRY_SEARCH || REGISTRY_KEY_EXISTS || REGISTRY_READ ||\
       //REGISTRY_WRITE || REGISTRY_READ_EXTRA || REGISTRY_WRITE_EXTRA ||\
       //REGISTRY_CREATE || REGISTRY_DELETE || REGISTRY_COPY_VALUE ||\
       //REGISTRY_COPY_KEY || REGISTRY_BACKUP

#if defined REGISTRY_SEARCH || defined REGISTRY_WRITE
int TypeNameToTypeValue(char *name)
{
	if (!lstrcmpi(name, "REG_BINARY")) return REG_BINARY;
	else if (!lstrcmpi(name, "REG_DWORD")) return REG_DWORD;
	else if (!lstrcmpi(name, "REG_DWORD_BIG_ENDIAN")) return REG_DWORD_BIG_ENDIAN;
	else if (!lstrcmpi(name, "REG_EXPAND_SZ")) return REG_EXPAND_SZ;
	else if (!lstrcmpi(name, "REG_MULTI_SZ")) return REG_MULTI_SZ;
	else if (!lstrcmpi(name, "REG_NONE")) return REG_NONE;
	else if (!lstrcmpi(name, "REG_SZ")) return REG_SZ;
	else if (!lstrcmpi(name, "REG_LINK")) return REG_LINK;
	else if (!lstrcmpi(name, "REG_RESOURCE_LIST")) return REG_RESOURCE_LIST;
	else if (!lstrcmpi(name, "REG_FULL_RESOURCE_DESCRIPTOR")) return REG_FULL_RESOURCE_DESCRIPTOR;
	else if (!lstrcmpi(name, "REG_RESOURCE_REQUIREMENTS_LIST")) return REG_RESOURCE_REQUIREMENTS_LIST;
	else if (!lstrcmpi(name, "REG_QWORD")) return REG_QWORD;
	else return -1;
}
#endif //REGISTRY_SEARCH || REGISTRY_WRITE

#if defined REGISTRY_SEARCH || defined REGISTRY_READ || defined REGISTRY_READ_EXTRA
char* TypeValueToTypeName(DWORD type)
{
	if (type == REG_BINARY) return "REG_BINARY";
	else if (type == REG_DWORD) return "REG_DWORD";
	else if (type == REG_DWORD_BIG_ENDIAN) return "REG_DWORD_BIG_ENDIAN";
	else if (type == REG_EXPAND_SZ) return "REG_EXPAND_SZ";
	else if (type == REG_MULTI_SZ) return "REG_MULTI_SZ";
	else if (type == REG_NONE) return "REG_NONE";
	else if (type == REG_SZ) return "REG_SZ";
	else if (type == REG_LINK) return "REG_LINK";
	else if (type == REG_RESOURCE_LIST) return "REG_RESOURCE_LIST";
	else if (type == REG_FULL_RESOURCE_DESCRIPTOR) return "REG_FULL_RESOURCE_DESCRIPTOR";
	else if (type == REG_RESOURCE_REQUIREMENTS_LIST) return "REG_RESOURCE_REQUIREMENTS_LIST";
	else if (type == REG_QWORD) return "REG_QWORD";
	else return "INVALID";
}

// Function: Converts registry data to string
void RegReadType(char *szOutput, char **pType, DWORD dwType, unsigned char *szString, DWORD dwSizeString)
{
	int a=0;
	int b=0;
	szOutput[0]='\0';

	if (dwType == REG_BINARY ||\
		dwType == REG_NONE ||\
		dwType == REG_LINK ||\
		dwType == REG_RESOURCE_REQUIREMENTS_LIST ||\
		dwType == REG_RESOURCE_LIST ||\
		dwType == REG_FULL_RESOURCE_DESCRIPTOR ||\
		dwType == REG_QWORD)
	{
		str2hex(szString, szOutput, TRUE, dwSizeString);
	}
	else if (dwType == REG_SZ ||\
		dwType == REG_EXPAND_SZ)
	{
		lstrcpyn(szOutput, (char *)szString, dwSizeString);
	}
	else if (dwType == REG_MULTI_SZ)
	{
		for (a=0, b=dwSizeString - 1; a < 2 && b >= 0 && !szString[b]; ++a, --b);

		for (a=0; a <= b; ++a)
		{
			if (!szString[a]) szOutput[a]='\n';
			else szOutput[a]=szString[a];
		}
		szOutput[a]='\0';
	}
	else if (dwType == REG_DWORD)
	{
		wsprintf(szOutput, "%d", *((unsigned int *)szString));
	}
	else if (dwType == REG_DWORD_BIG_ENDIAN)
	{
		wsprintf(szOutput, "%d", SWAP_ENDIAN(*((unsigned int *)szString)));
	}

	*pType=TypeValueToTypeName(dwType);
}
#endif //REGISTRY_SEARCH || REGISTRY_READ || REGISTRY_READ_EXTRA

#if defined REGISTRY_WRITE || defined REGISTRY_WRITE_EXTRA
// Function from NSIS source: Converts string with binary data to binary data
void HexStrToHexData(char *pHexStr, unsigned char *pHexData, int *nHexLen)
{
	int nLen=0;
	int a,b;
	int c;

	while (*pHexStr)
	{
		a=*pHexStr;
		if (a >= '0' && a <= '9') a-='0';
		else if (a >= 'a' && a <= 'f') a-='a'-10;
		else if (a >= 'A' && a <= 'F') a-='A'-10;
		else break;
		b=*++pHexStr;
		if (b >= '0' && b <= '9') b-='0';
		else if (b >= 'a' && b <= 'f') b-='a'-10;
		else if (b >= 'A' && b <= 'F') b-='A'-10;
		else break;
		pHexStr++;
		c=(a<<4)|b;
		pHexData[nLen++]=c;
	}
	*nHexLen=nLen;
}
#endif //REGISTRY_WRITE || REGISTRY_WRITE_EXTRA

#if defined REGISTRY_DELETE || defined REGISTRY_COPY_KEY
LONG RegDeleteSubKeys(HKEY hKey, LPCTSTR lpSubKey)
{
	char szKeyTmp[MAX_PATHLEN];
	HKEY hSubKey;
	DWORD dwSizeKey=sizeof(szKeyTmp);
	DWORD dwIndex=0;
	int nError;

	if ((nError=API_RegOpenKeyEx(hKey, lpSubKey, 0, KEY_ENUMERATE_SUB_KEYS, &hSubKey)) == ERROR_SUCCESS)
	{
		while (API_RegEnumKeyEx(hSubKey, dwIndex, szKeyTmp, &dwSizeKey, NULL, NULL, NULL, NULL) == ERROR_SUCCESS)
		{
			if (RegDeleteSubKeys(hSubKey, szKeyTmp) != ERROR_SUCCESS)
				++dwIndex;
			dwSizeKey=sizeof(szKeyTmp);
		}
		API_RegCloseKey(hSubKey);
		nError=API_RegDeleteKey(hKey, lpSubKey);
	}
	return nError;
}

// Function: Tries to delete registry key in one step (Win9x/Me),
// if error then calls RegDeleteSubKeys
int RegDeleteKeyPath(HKEY hKeyRoot, char *szPath)
{
	char *pKeyName=szPath + lstrlen(szPath) - 1;
	HKEY hKeyHandle;
	int nError=-1;

	while (pKeyName >= szPath && *pKeyName != '\\') --pKeyName;

	if (pKeyName < szPath) return -1;
	*pKeyName++='\0';

	if (API_RegOpenKeyEx(hKeyRoot, szPath, 0, KEY_ENUMERATE_SUB_KEYS, &hKeyHandle) == ERROR_SUCCESS)
	{
		if ((nError=API_RegDeleteKey(hKeyHandle, pKeyName)) != ERROR_SUCCESS)
		{
			nError=RegDeleteSubKeys(hKeyHandle, pKeyName);
		}
		API_RegCloseKey(hKeyHandle);
	}
	return (nError == ERROR_SUCCESS)?0:-1;
}
#endif //REGISTRY_DELETE || REGISTRY_COPY_KEY

#if defined REGISTRY_COPY_KEY
// Function: Recursively copies registry key
LONG RegCopyKey(HKEY hSource, char *szPathSource, HKEY hTarget, char *szPathTarget)
{
	//Global variables: szKey, szValue, szString

	DWORD dwSizeKey;
	DWORD dwSizeValue;
	DWORD dwSizeString;
	DWORD dwType;
	DWORD dwIndex;
	DWORD dwDisposition;
	HKEY hSourceHandle;
	HKEY hTargetHandle;
	int nError;

	if ((nError=API_RegOpenKeyEx(hSource, szPathSource, 0, KEY_READ, &hSourceHandle)) != ERROR_SUCCESS ||\
		(nError=API_RegCreateKeyEx(hTarget, szPathTarget, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hTargetHandle, &dwDisposition)) != ERROR_SUCCESS)
	{
		return nError;
	}

	dwIndex=0;
	while (1)
	{
		dwSizeValue=sizeof(szValue);
		dwSizeString=sizeof(szString);
		szString[0]='\0';
		if (API_RegEnumValue(hSourceHandle, dwIndex++, szValue, &dwSizeValue, NULL, &dwType, szString, &dwSizeString) != ERROR_SUCCESS)
			break;

		API_RegSetValueEx(hTargetHandle, szValue, 0, dwType, szString, dwSizeString);
	}

	dwIndex=0;
	while (1)
	{
		dwSizeKey=sizeof(szKey);
		if (API_RegEnumKeyEx(hSourceHandle, dwIndex++, szKey, &dwSizeKey, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
		{
			API_RegCloseKey(hSourceHandle);
			API_RegCloseKey(hTargetHandle);
			return ERROR_SUCCESS;
		}

		RegCopyKey(hSourceHandle, szKey, hTargetHandle, szKey);
	}
}
#endif //REGISTRY_COPY_KEY

#ifdef REGISTRY_BACKUP
// Function: Writes registry data to file (REGEDIT4 format)
void RegWriteType(HANDLE hFileWrite, char *szValueName, DWORD dwType, unsigned char *szString, DWORD dwSizeString)
{
	//Global variables: szBuf, szBuf2

	unsigned int i=0;
	int nCurWidth=0;
	int nColumnWidth;
	DWORD dwNumberOfBytesWritten;

	if (!*szValueName)
		lstrcpy(szBuf2, "@");
	else
	{
		ReplaceEscapeSequence(szBuf, szValueName);
		wsprintf(szBuf2, "\"%s\"", szBuf);
	}

	if (dwType == REG_BINARY ||\
		dwType == REG_EXPAND_SZ ||\
		dwType == REG_MULTI_SZ ||\
		dwType == REG_DWORD_BIG_ENDIAN ||\
		dwType == REG_NONE ||\
		dwType == REG_LINK ||\
		dwType == REG_RESOURCE_LIST ||\
		dwType == REG_FULL_RESOURCE_DESCRIPTOR ||\
		dwType == REG_RESOURCE_REQUIREMENTS_LIST ||\
		dwType == REG_QWORD)
	{
		if (dwType == REG_BINARY)
			lstrcat(szBuf2, "=hex:");
		else if (dwType == REG_NONE)
			lstrcat(szBuf2, "=hex(0):");
		else if (dwType == REG_EXPAND_SZ)
			lstrcat(szBuf2, "=hex(2):");
		else if (dwType == REG_DWORD_BIG_ENDIAN)
			lstrcat(szBuf2, "=hex(5):");
		else if (dwType == REG_LINK)
			lstrcat(szBuf2, "=hex(6):");
		else if (dwType == REG_MULTI_SZ)
			lstrcat(szBuf2, "=hex(7):");
		else if (dwType == REG_RESOURCE_LIST)
			lstrcat(szBuf2, "=hex(8):");
		else if (dwType == REG_FULL_RESOURCE_DESCRIPTOR)
			lstrcat(szBuf2, "=hex(9):");
		else if (dwType == REG_RESOURCE_REQUIREMENTS_LIST)
			lstrcat(szBuf2, "=hex(a):");
		else if (dwType == REG_QWORD)
			lstrcat(szBuf2, "=hex(b):");

		if (dwSizeString == 0)
		{
			lstrcat(szBuf2, "\r\n");
			WriteFile(hFileWrite, szBuf2, lstrlen(szBuf2), &dwNumberOfBytesWritten, NULL);
			return;
		}

		nColumnWidth=(79 - lstrlen(szBuf2)) / 3;

		if (nColumnWidth <= 0)
			nColumnWidth=1;

		for (; i < dwSizeString; lstrcpy(szBuf2, "  "), nColumnWidth=25)
		{
			for (nCurWidth=0; i < dwSizeString && nCurWidth < nColumnWidth; ++i, ++nCurWidth)
			{
				wsprintf(szName, "%02x,", (unsigned int)szString[i]);
				lstrcat(szBuf2, szName);
			}
			if (i == dwSizeString)
			{
				szBuf2[lstrlen(szBuf2) - 1]='\0';
				lstrcat(szBuf2, "\r\n");
			}
			else lstrcat(szBuf2, "\\\r\n");

			WriteFile(hFileWrite, szBuf2, lstrlen(szBuf2), &dwNumberOfBytesWritten, NULL);
		}
	}
	else if (dwType == REG_SZ)
	{
		ReplaceEscapeSequence(szBuf, szString);
		lstrcat(szBuf2, "=\"");
		lstrcat(szBuf2, szBuf);
		lstrcat(szBuf2, "\"\r\n");
		WriteFile(hFileWrite, szBuf2, lstrlen(szBuf2), &dwNumberOfBytesWritten, NULL);
	}
	else if (dwType == REG_DWORD)
	{
		wsprintf(szBuf, "%s=dword:%08x\r\n", szBuf2, *((unsigned int *)szString));
		WriteFile(hFileWrite, szBuf, lstrlen(szBuf), &dwNumberOfBytesWritten, NULL);
	}
}

// Function: Converts escape sequences (\\ -> \\\\), (\" -> \\\"), (\r -> \\r), (\n -> \\n)
void ReplaceEscapeSequence(char *pResult, char *pText)
{
	int a=0;
	int b=0;

	for (; (pText[a]); ++a, ++b)
		if (pText[a] == '\\')
			pResult[b]='\\', pResult[++b]='\\';
		else if (pText[a] == '\"')
			pResult[b]='\\', pResult[++b]='\"';
		else if (pText[a] == '\r')
			pResult[b]='\\', pResult[++b]='r';
		else if (pText[a] == '\n')
			pResult[b]='\\', pResult[++b]='n';
		else pResult[b]=pText[a];
	pResult[b]='\0';
}

#ifndef POCKETPC
// Function: Checks is file exist
BOOL FileExists(char *fname)
{
	WIN32_FIND_DATA wfd;
	HANDLE hFind=FindFirstFile(fname, &wfd);
	if (INVALID_HANDLE_VALUE != hFind)
	{
		FindClose(hFind);
		return TRUE;
	}
	return FALSE;
}
#endif //POCKETPC
#endif //REGISTRY_BACKUP


//Standart registry API calls
LONG API_RegOpenKeyEx(HKEY hKey, LPCTSTR lpSubKey, DWORD ulOptions, REGSAM c, PHKEY phkResult)
{
	LONG nError;
#ifndef POCKETPC
	nError=RegOpenKeyEx(hKey, lpSubKey, 0, c, phkResult);
#else
	MultiByteToWideChar(CP_ACP, 0, lpSubKey, -1, wszBuf, MAX_PATHLEN);
	nError=CeRegOpenKeyEx(hKey, wszBuf, 0, 0, phkResult);
#endif
	return nError;
}

LONG API_RegEnumKeyEx(HKEY hKey, DWORD dwIndex, LPTSTR lpName, LPDWORD lpcName, LPDWORD lpReserved, LPTSTR lpClass, LPDWORD lpcClass, PFILETIME lpftLastWriteTime)
{
	LONG nError;
#ifndef POCKETPC
	nError=RegEnumKeyEx(hKey, dwIndex, lpName, lpcName, NULL, NULL, NULL, NULL);
#else
	nError=CeRegEnumKeyEx(hKey, dwIndex, wszBuf, lpcName, NULL, NULL, NULL, NULL);

	if (nError == ERROR_SUCCESS)
	{
		WideCharToMultiByte(CP_ACP, 0, wszBuf, -1, lpName, MAX_PATHLEN, NULL, NULL);
	}
#endif
	return nError;
}

LONG API_RegCreateKeyEx(HKEY hKey, LPCTSTR lpSubKey, DWORD Reserved, LPTSTR lpClass, DWORD dwOptions, REGSAM samDesired, LPSECURITY_ATTRIBUTES lpSecurityAttributes, PHKEY phkResult, LPDWORD lpdwDisposition)
{
	LONG nError;
#ifndef POCKETPC
	nError=RegCreateKeyEx(hKey, lpSubKey, 0, NULL, dwOptions, samDesired, NULL, phkResult, lpdwDisposition);
#else
	MultiByteToWideChar(CP_ACP, 0, lpSubKey, -1, wszBuf, MAX_PATHLEN);
	nError=CeRegCreateKeyEx(hKey, wszBuf, 0, L"", 0, 0, NULL, phkResult, lpdwDisposition);
#endif
	return nError;
}

LONG API_RegCloseKey(HKEY hKey)
{
	LONG nError;
#ifndef POCKETPC
	nError=RegCloseKey(hKey);
#else
	nError=CeRegCloseKey(hKey);
#endif
	return nError;
}

LONG API_RegDeleteKey(HKEY hKey, LPCTSTR lpSubKey)
{
	LONG nError;
#ifndef POCKETPC
	nError=RegDeleteKey(hKey, lpSubKey);
#else
	MultiByteToWideChar(CP_ACP, 0, lpSubKey, -1, wszBuf, MAX_PATHLEN);
	nError=CeRegDeleteKey(hKey, wszBuf);
#endif
	return nError;
}

LONG API_RegEnumValue(HKEY hKey, DWORD dwIndex, LPTSTR lpValueName, LPDWORD lpcValueName, LPDWORD lpReserved, LPDWORD lpType, LPBYTE lpData, LPDWORD lpcbData)
{
	LONG nError;
#ifndef POCKETPC
	nError=RegEnumValue(hKey, dwIndex, lpValueName, lpcValueName, NULL, lpType, lpData, lpcbData);
#else
	unsigned char *pwszData=(unsigned char *)wszData;
	DWORD i;

	if (!lpData) pwszData=NULL;
	nError=CeRegEnumValue(hKey, dwIndex, wszBuf, lpcValueName, NULL, lpType, pwszData, lpcbData);

	if (nError == ERROR_SUCCESS)
	{
		WideCharToMultiByte(CP_ACP, 0, wszBuf, -1, lpValueName, MAX_PATHLEN, NULL, NULL);

		if (pwszData)
		{
			if (*lpType == REG_SZ ||\
				*lpType == REG_EXPAND_SZ)
			{
				WideCharToMultiByte(CP_ACP, 0, wszData, -1, lpData, MAX_DATALEN, NULL, NULL);
				*lpcbData/=sizeof(wchar_t);
			}
			else for (i=0; i <= *lpcbData; ++i) lpData[i]=pwszData[i];
		}
	}
#endif
	return nError;
}

LONG API_RegDeleteValue(HKEY hKey, LPCTSTR lpValueName)
{
	LONG nError;
#ifndef POCKETPC
	nError=RegDeleteValue(hKey, lpValueName);
#else
	MultiByteToWideChar(CP_ACP, 0, lpValueName, -1, wszBuf, MAX_PATHLEN);
	nError=CeRegDeleteValue(hKey, wszBuf);
#endif
	return nError;
}

LONG API_RegQueryValueEx(HKEY hKey, LPCTSTR lpValueName, LPDWORD lpReserved, LPDWORD lpType, LPBYTE lpData, LPDWORD lpcbData)
{
	LONG nError;
#ifndef POCKETPC
	nError=RegQueryValueEx(hKey, lpValueName, NULL, lpType, lpData, lpcbData);
#else
	unsigned char *pwszData=(unsigned char *)wszData;
	DWORD i;

	if (!lpData) pwszData=NULL;
	MultiByteToWideChar(CP_ACP, 0, lpValueName, -1, wszBuf, MAX_PATHLEN);
	nError=CeRegQueryValueEx(hKey, wszBuf, NULL, lpType, pwszData, lpcbData);

	if (nError == ERROR_SUCCESS)
	{
		if (pwszData)
		{
			if (*lpType == REG_SZ ||\
				*lpType == REG_EXPAND_SZ)
			{
				WideCharToMultiByte(CP_ACP, 0, wszData, -1, lpData, MAX_DATALEN, NULL, NULL);
				*lpcbData/=sizeof(wchar_t);
			}
			else for (i=0; i <= *lpcbData; ++i) lpData[i]=pwszData[i];
		}
	}
#endif
	return nError;
}

LONG API_RegSetValueEx(HKEY hKey, LPCTSTR lpValueName, DWORD Reserved, DWORD dwType, BYTE* lpData, DWORD cbData)
{
	LONG nError;
#ifndef POCKETPC
	nError=RegSetValueEx(hKey, lpValueName, 0, dwType, lpData, cbData);
#else
	unsigned char *pwszData=(unsigned char *)wszData;

	MultiByteToWideChar(CP_ACP, 0, lpValueName, -1, wszBuf, MAX_PATHLEN);

	if (dwType == REG_SZ ||\
		dwType == REG_EXPAND_SZ)
	{
		MultiByteToWideChar(CP_ACP, 0, lpData, -1, wszData, MAX_DATALEN);
		cbData*=sizeof(wchar_t);
	}
	else pwszData=lpData;

	nError=CeRegSetValueEx(hKey, wszBuf, 0, dwType, pwszData, cbData);
#endif
	return nError;
}


#if defined REGISTRY_WRITE || defined REGISTRY_READ_EXTRA
int popinteger(char *str)
{
	popstring(str);
	return xatoi(str);
}
#endif //REGISTRY_WRITE || REGISTRY_READ_EXTRA

//Function: Removes the element from the top of the NSIS stack and puts it in the buffer
int popstring(char *str)
{
	stack_t *th;
	if (!g_stacktop || !*g_stacktop) return 1;
	th=(*g_stacktop);
	lstrcpy(str,th->text);
	*g_stacktop = th->next;
	GlobalFree((HGLOBAL)th);
	return 0;
}

//Function: Adds an element to the top of the NSIS stack
void pushstring(const char *str)
{
	stack_t *th;
	if (!g_stacktop) return;
	th=(stack_t*)GlobalAlloc(GPTR,sizeof(stack_t)+g_stringsize);
	lstrcpyn(th->text,str,g_stringsize);
	th->next=*g_stacktop;
	*g_stacktop=th;
}
